/*------------------------------------------------------------------*/
/* Module:      Used.c                                              */
/* Description: determines bytes used and bytes allocated by all    */
/*              files in a given directory and its subdirectories   */
/* Author:      Paul Bruno - Copyright 1988 SomeWare                */
/* Compiling:   This module has been compiled with MS 'C' versions  */
/*              4.0 and 5.1, without modification.  The "/AL" and   */
/*              "/Zp" options must be used. "/Ox" and "/Oa" CANNOT  */
/*              be used with confidence.  "/Ot" and "/Gs" are       */
/*              recommended.                                        */
/* Linking:     "/EXEPACK" is recommended.                          */
/*                                                                  */
/*------------------------------------------------------------------*/
#include <dos.h>
#include <stdio.h>
#include <memory.h>
#include <malloc.h>
#include <ctype.h>

#define BOOL            short
#define TRUE            1
#define FALSE           0
#define HEADER_STR      "\n\
Directory                    Files  Bytes Used  Bytes Allocated  % Wasted\n\
---------                    -----  ----------  ---------------  --------"


struct DREN_tp {             /* must be packed with Zp */
   unsigned char Reserved[21];
   unsigned char AttrByte;
   unsigned char FTime[2];
   unsigned char FDate[2];
            long FSz;
   unsigned char FSpec[13];
   };

struct SDIR_tp {              /* internal storage for subdirs */
   short FrstSubDir;
   short NxtSubDir;
   short PrntDir;
   char  FllPth[42];
   };
long BytsInDir;
long BytsAllctd;
long TtlAllctd = 0;
long TtlUsed   = 0;
short FlsInDir;
long  TtlFls = 0;
short DirToChk, NxtDirAvl;
short ClusterSz, RndUpVal;
char *DTA;
char  *FlsToGt;
short CurLn = 1;
struct DREN_tp *DREN_pt;
struct SDIR_tp *SDIR_ar;
short LnsPerPg = 54;
BOOL Pause = FALSE;
BOOL ChkSubDirs = TRUE;

void StDTA(NewDTA)
   char *NewDTA;
{
   union REGS in_regs, out_regs;
   struct SREGS seg_regs;

   in_regs.h.ah = 0x01A;   /* set DTA */
   in_regs.x.dx = FP_OFF(NewDTA);
   seg_regs.ds  = FP_SEG(NewDTA);

   intdosx(&in_regs, &out_regs, &seg_regs);
   }

short DOSFndFrst(path)
   char *path;
{
   union REGS in_regs, out_regs;
   struct SREGS seg_regs;

   in_regs.h.ah = 0x04E;  /* fnd frst */
   in_regs.x.cx = 0x037;  /* fnd all types but vol label */
   in_regs.x.dx = FP_OFF(path);
   seg_regs.ds  = FP_SEG(path);
   intdosx(&in_regs, &out_regs, &seg_regs);

   if (out_regs.x.cflag)
      return(out_regs.x.ax);
   else
      return(0);  /* no error */
   }

BOOL DOSFndNxt()
{
   union REGS in_regs, out_regs;

   in_regs.h.ah = 0x04F;  /* fnd next */
   intdos(&in_regs, &out_regs);

   return((out_regs.x.cflag) ? FALSE : TRUE);
   }

BOOL AllcGlbls()
{
   FlsToGt = _fmalloc(80);
   if (!(DTA = _fmalloc(2048)))
      return(FALSE);
   DREN_pt = (struct DREN_tp *) DTA;

   if (!(SDIR_ar = (struct SDIR_tp *) _fmalloc(512 * sizeof(struct SDIR_tp))))
      return(FALSE);
   return(TRUE);
   }

void ShowTotals()
{
   if (++CurLn >= LnsPerPg) {
      if (Pause) {
         fprintf(stderr, "Press any key to continue");
         getch();
         fputc('\n', stderr);
         }
      putch('\f');
      puts(HEADER_STR);
      CurLn = 3;
      }
   printf("%-28s %5d  %10ld  %15ld     %2d\n", SDIR_ar[DirToChk].FllPth,
    FlsInDir, BytsInDir, BytsAllctd, (BytsAllctd) ? 
    (short) ((float) ((float) (BytsAllctd - BytsInDir)/BytsAllctd) * 100) : 0);

   }

BOOL GtNxtToChk()
{
   if (! ChkSubDirs)
      return(FALSE);
   while ((SDIR_ar[DirToChk].NxtSubDir == -1) || 
    (SDIR_ar[(SDIR_ar[DirToChk].NxtSubDir)].PrntDir != DirToChk)) {
      if (SDIR_ar[DirToChk].NxtSubDir != -1)
         NxtDirAvl = SDIR_ar[DirToChk].FrstSubDir;
      if ((DirToChk = SDIR_ar[DirToChk].PrntDir) == -1)
         return(FALSE);
      }
   DirToChk = (SDIR_ar[DirToChk].NxtSubDir)++;
   if ((DirToChk >= NxtDirAvl) || (DirToChk == -1))
      return(FALSE);
   return(TRUE);
   }
BOOL GtDskInfo(path)
   char *path;
{
   union REGS in_regs, out_regs;

   if (path[1] == ':') {
      path[0] = toupper(path[0]);
      in_regs.h.dl = path[0] - ('A' - 1);
      }
   else
      in_regs.h.dl = 0;  /* get default */
   in_regs.h.ah = 0x036;
   intdos(&in_regs, &out_regs);
   if (out_regs.x.ax == 0x0FFFF)
      return(FALSE);
   ClusterSz = out_regs.x.ax * out_regs.x.cx;
   RndUpVal = ClusterSz - 1;
   return(TRUE);
   }

ChkNxtEnt()
{
   struct SDIR_tp *SDIR_pt;

   if (DREN_pt -> AttrByte & 0x010) {
      if(strcmp(DREN_pt -> FSpec, ".") && strcmp(DREN_pt -> FSpec, "..")) {
         if (SDIR_ar[DirToChk].FrstSubDir == -1) {
            SDIR_ar[DirToChk].FrstSubDir = (SDIR_ar[DirToChk].NxtSubDir =
             NxtDirAvl);
            }
         SDIR_pt = &SDIR_ar[NxtDirAvl++];
         sprintf(SDIR_pt -> FllPth, "%s%s\\", SDIR_ar[DirToChk].FllPth,
          DREN_pt -> FSpec);
         SDIR_pt -> NxtSubDir = (SDIR_pt -> FrstSubDir = -1);
         SDIR_pt -> PrntDir = DirToChk;
         }
      }
   else {
      BytsInDir += DREN_pt -> FSz;
      BytsAllctd += (((DREN_pt -> FSz) + RndUpVal) / ClusterSz) * ClusterSz;
      FlsInDir++;
      }
   }
ShowFinalTtls()
{
   printf("\n%6ld files containing %9ld bytes have been allocated %9ld bytes",
    TtlFls, TtlUsed, TtlAllctd);
   printf("\n%6d %% of allocated space is unused.\n", (TtlAllctd) ?
    (short) ((float) ((float) (TtlAllctd - TtlUsed)/TtlAllctd) * 100) : 0);
   }
void GvHlp()
{
   puts("\nUsage:\n\n   used [-p] [-l=<lines>] [-n] [<path to check>]");
   puts("\nInput:\n\n   -p pauses the output after each page");
   puts("   -l=<lines> sets the number of lines per page (default is 54)");
   puts("   -n indicates that subdirectories will not be checked");
   puts("   <path to check> is the directory to check (default is current)");
   }

BOOL PrcssSwtch(arg)
   char *arg;
{
   switch (toupper(arg[0])) {
      case 'N':
         ChkSubDirs = FALSE;
         break;
      case 'P':
         Pause = TRUE;
         break;
      case 'L':
         if (arg[1] == '=') {
            if ((LnsPerPg = atoi(arg + 2)) <= 0)
               LnsPerPg = 54;
            }
         else {
            GvHlp();
            return(FALSE);
            }
         break;
      default:
         GvHlp();
         return(FALSE);
         break;
      }
   return(TRUE);
   }

BOOL GtCurDir(path_buf)
   char *path_buf;
{
   union REGS in_regs, out_regs;
   struct SREGS seg_regs;

   in_regs.h.ah = 0x47;                     /* get cur directory */
   in_regs.h.dl = 0;                        /* for current drive */
   in_regs.x.si = FP_OFF(path_buf);
   seg_regs.ds = FP_SEG(path_buf);
   intdosx(&in_regs, &out_regs, &seg_regs);

   return((out_regs.x.cflag) ? FALSE : TRUE);
   }

BOOL GtArgs(argc, argv)
   int argc;
   char **argv;
{
   short nxt_arg = 1;

   while ((argc > nxt_arg) &&
    ((argv[nxt_arg][0] == '-') || (argv[nxt_arg][0] == '/'))) {
      if (!PrcssSwtch(&argv[nxt_arg++][1]))
         return(FALSE);
      }
   if (argc > nxt_arg)
      strcpy(FlsToGt,argv[nxt_arg]);
   else {
      FlsToGt[0] = '\\';
      if (!GtCurDir(FlsToGt+1)) {
         perror("Error getting current directory");
         GvHlp();
         return(FALSE);
         }
      }
   return(TRUE);
   }
void InitGlbls()
{
   short pth_sz;
   short i;
   pth_sz = strlen(FlsToGt);
   if ((FlsToGt[pth_sz - 1] != '\\') || (!pth_sz)) {
      FlsToGt[pth_sz] = '\\';
      FlsToGt[pth_sz + 1] = '\0';
      }

   DirToChk = 0;
   NxtDirAvl = 1;
   memset((char *) SDIR_ar, (char) -1,
    (unsigned) (512 * sizeof(struct SDIR_tp)));
   strcpy(SDIR_ar[0].FllPth, FlsToGt);
   }

main(argc, argv)
   int argc;
   char **argv;
{
   short  pth_sz;

   puts("\nUSED - Copyright (C) SomeWare 1988, by Paul Bruno - All Rights\
 Reserved");

   if (!AllcGlbls()) {
      perror("Error Allocating Global Data");
      GvHlp();
      exit(1);
      }

   StDTA(DTA);
   if (!GtArgs(argc, argv))
      exit(1);

   if (!GtDskInfo(FlsToGt)) {
      perror("Error getting drive information");
      GvHlp();
      exit(1);
      }

   InitGlbls();

   puts(HEADER_STR);
   CurLn = 4;
   do {
      sprintf(FlsToGt, "%s*.*", SDIR_ar[DirToChk].FllPth);

      switch(DOSFndFrst(FlsToGt)) {
         case 2:
         default:   /* not all return values documented */
            perror("Invalid path specification");
            GvHlp();
            exit(1);
            break;
         case 0x012:    /* no matching entry */
            ShowTotals();
            ShowFinalTtls();
            exit(0);
            break;
         case 0:  /* found one */
            break;
         }
      BytsAllctd = (BytsInDir = (FlsInDir = 0));
      do {
         ChkNxtEnt();
         }
         while(DOSFndNxt());
      ShowTotals();
      TtlAllctd += BytsAllctd;
      TtlUsed   += BytsInDir;
      TtlFls    += FlsInDir;
      }
      while (GtNxtToChk());
   ShowFinalTtls();
   exit(0);
   }
